#include "CPP.H"
#include "Interface.H"
#include "PlayLists.H"
#include <boost/format.hpp>
#include <boost/algorithm/string.hpp>
#include "MPD.H"

using namespace std;
using namespace boost::filesystem;

vector<string> Interface::mode_names = {
  "Audio",
  "Video",
  "Blu",
  "Wii",
  "CD-Classic",
  "CD-Pop",
  "Off"
};


Interface::Interface(const std::string &music_dir,
                     ostream &vfd_os_,
                     mpd_connection *mpd_conn,
                     RNG &rng,
                     int update_hz,
                     Mode initial_mode,
                     int play_mode)
  : vfd_os(vfd_os_), mpd_conn(mpd_conn), rng(rng), update_hz(update_hz), mode(initial_mode), play_mode(play_mode)
{
  shuffle = true;
  repeat  = true;
  playing = false;
  pausing = false;
  mode2 = false;
  classify_index = 0;

  home_path = path(music_dir);
  if (!exists(home_path) || !is_directory(home_path))
    ERR(music_dir << " not a directory");

  scan_files();
}


std::string Interface::status() const
{
  string text = mode_names[mode] + " ";

  if (playing) {

    if (pausing) text += "Pausing";
    else         text += "Playing";

    text += (boost::format(" %d/%d %d:%02d/%d:%02d %d-kbits :\n")
             % (mpd_status.song_pos+1)
             % mpd_status.queue_len
             % (mpd_status.elapsed_time / 60)
             % (mpd_status.elapsed_time % 60)
             % (mpd_status.total_time / 60)
             % (mpd_status.total_time % 60)
             % mpd_status.bit_rate).str();
    
    int sp = mpd_status.song_pos;
    if (sp >= 0 && sp < (int)songs.size()) {
      string song = songs[mpd_status.song_pos];
      boost::replace_all(song, ".mp3", "");
      text += song;
    }
    
  } else {

    if (mode2) {

      // play mode selection

      text += (boost::format("Play Mode %d/%d:\n%s")
               % (play_mode.get_state()+1)
               % play_mode.NUM_STATES
               % play_mode.to_string()).str();

    } else {

      string wp;
      int ci;
    
      if (work_item_nodes.size()) {
        wp = work_item_strs[current_item];
        ci = current_item + 1;
      } else {
        wp = work_node->path;
        ci = 0;
      }
    
      // track/total
      text += (boost::format("Selection %d/%d :\n%s")
               % ci
               % work_item_nodes.size()
               % wp
               ).str();
    }
  }
  return text;
}


void Interface::scan_files()
{
  vfd_os << "Scan and mark files ..." << flush;
  delete play_node_tree;
  play_node_tree = PlayNode::create_tree(home_path, home_path);
  play_node_tree->mark(play_mode);

  item_stack.clear();
  work_node_stack.clear();
  
  new_work_node(play_node_tree, 0);
}


// update

void Interface::update(int tick_)
{
  tick = tick_;

  if (tick % (update_hz/2) == 0 && mpd_conn) {

    // get MPD status ~ twice per second

    mpd_status.update(mpd_conn);
    mpd_status_updated = true;

    if (playing) {
      if (mpd_status.song_pos < 0) {

        // list played completely
        playing = false;

        cout << "REPEAT" << endl;
        
        if (repeat)
          action_playpause();
      }
    }
    
    //mpd_status.print();
    //cout << endl;
  }

  vfd_update();
}


// actions

void Interface::action_select()
{
  last_action_tick = tick;

  if (playing) return;

  selection_tick = tick;

  if (mode2) {

    // file refresh and back
    scan_files();
    mode2 = 0;

  } else {

    if (work_item_nodes.empty()) return;
    if (!work_item_nodes[current_item]) return; // *
    
    if (work_item_nodes[current_item]->directory) {
      item_stack.push_back(current_item);
      work_node_stack.push_back(work_node);
      new_work_node(work_item_nodes[current_item], 0);
    }
  }
}

void Interface::action_cw()
{
  last_action_tick = tick;
  
  if (playing) return;

  selection_tick = tick;
  
  if (mode2) {

    play_mode.increment();
    
  } else {

    current_item++;  
    if (current_item >= (int)work_item_nodes.size())
      current_item = 0;
  }
  
  vfd_update();
}

void Interface::action_ccw()
{
  last_action_tick = tick;

  if (playing) return;

  selection_tick = tick;

  if (mode2) {

    play_mode.decrement();
    
  } else {
  
    current_item--;
    if (current_item < 0)
      current_item = (int)work_item_nodes.size() - 1;
  }
  
  vfd_update();  
}

void Interface::action_previous()
{
  last_action_tick = tick;  

  if (playing) {
    freeze_ticks = FREEZE_SECS*update_hz;
    delta_tick1 = tick; // title start
    mpd_previous(mpd_conn);
    vfd_update();
    return;
  }

  selection_tick = tick;

  if (mode2) return;
  
  if (work_node != play_node_tree) {

    // up one folder
    new_work_node(work_node_stack.back(), item_stack.back());
    item_stack.pop_back();
    work_node_stack.pop_back();

  } else {

    // refresh tree if root "previou" in root node
    scan_files();
  }
}

void Interface::action_next()
{
  last_action_tick = tick;

  if (playing) {
  
    freeze_ticks = FREEZE_SECS*update_hz;
    delta_tick1 = tick; // title start  
    mpd_next(mpd_conn);
    vfd_update();
    return;
  }
}

void Interface::action_seek_backward()
{
  seek(SEEK_BACKWARD_SECS);
}

void Interface::action_seek_forward()
{
  seek(SEEK_FORWARD_SECS);
}

// helper
void Interface::seek(int secs)
{
  last_action_tick = tick;  

  if (!playing) return;

  freeze_ticks = FREEZE_SECS*update_hz;
  delta_tick1 = tick; // title start

  mpd_status.update(mpd_conn);
  mpd_status_updated = true;

  int seek_time = mpd_status.elapsed_time + secs;
  if (seek_time >= mpd_status.total_time)
    seek_time = mpd_status.total_time - 10;
  if (seek_time < 0)
    seek_time = 0;
  
  mpd_seek(mpd_conn, mpd_status.song_pos, seek_time);
  vfd_update();
}

void Interface::action_stop()
{
  last_action_tick = tick;  
  playing = false;
  pausing = false;
  mode2 = false;
  mpd_stop(mpd_conn);
}

void Interface::action_mode()
{
  last_action_tick = tick;  

  if (playing) return;

  if (mode2) {
    
    // file refresh and back
    scan_files();
    mode2 = false;

  } else {

    // switch to mode selection
    mode2 = true;
  }

  vfd_update();
}

void Interface::action_playpause()
{
  last_action_tick = tick;
  
  if (work_item_nodes.empty())
    return;
  
  if (playing) {

    pausing = !pausing;
    mpd_pause(mpd_conn, pausing);

  } else {

    playing = true;
    pausing = false;
    mode2 = false;
    
    mpd_clear(mpd_conn);
    
    PlayLists lists("");
    vector<string> list;

    if (!work_item_nodes[current_item]) {
      // * => all files/directories
      FORALL (work_item_nodes, it) {
        if (*it) {
          list.push_back((*it)->path);
        }
      }
      
    } else {
      list.push_back(work_item_nodes[current_item]->path);
    }
    
    cout << "List: " << list.back() << endl;
    cout << "home: " << home_path.string() << endl;
    lists.add_list("playlist", list);
    string display_name;
    lists.create_song_list(0, play_node_tree, rng, shuffle, play_mode, display_name, songs);

    // prune long lists
    if (songs.size() > 200) {
      songs.erase(songs.begin() + 200, songs.end());
    }

    mpd_add(mpd_conn, songs);
    mpd_play(mpd_conn);
  }
}

void Interface::action_toggle_shuffle()
{
  last_action_tick = tick;  
  shuffle = !shuffle;
}

void Interface::vfd(const string &msg)
{
  string m = msg;
  if (m.size() > 32)
    m = m.substr(0, 32);
  while (m.size() < 32) {
    m += ' ';
  }

  if (&vfd_os == &cout) {
    vfd_os << m.substr(0, 16) << endl;
    if (m.size() >= 16) 
      vfd_os << m.substr(16) << endl;
  } else {
    cout << "Send to VFD: " << m << endl;
    vfd_os << m << flush;
  }
}

void Interface::new_work_node(PlayNode *node, int new_item)
{
  work_node = node;
  work_item_strs.clear();
  work_item_nodes.clear();
  
  int max_length = -1;

  FORALL (work_node->succ, it) {
    if ((*it)->marked) {
      string ws = (*it)->path;
      work_item_strs.push_back(ws);
      work_item_nodes.push_back(*it);
      max_length = ::max(max_length, (int)ws.size());
    }
  }

  if (work_node == play_node_tree && work_item_nodes.size() > 1) {
    // * entry at root => play everything
    work_item_strs.push_back("*");
    work_item_nodes.push_back(nullptr);
  }

  current_item = new_item;
  int n = work_item_nodes.size();
  if (new_item >= n)
    new_item = n - 1;

  if (new_item < 0)
    new_item = 0;
  
  // compute starting point for displaying paths
  // if one string is longer than 24, only show
  // 2 bottom levels

  string wp_s = work_node->path;
  int wp_l = wp_s.size();
  work_path_start_index = 0;
  
  if (max_length < 0)
    max_length = wp_l; // empty dir, only display path

  // "x/x " "xx/xx " "xxx/xxx "
  int prefix_length = 4;
  if (n > 9) {
    prefix_length += 2;
    if (n > 99) {
      prefix_length += 2;
    }
  }
  
  if (max_length + prefix_length >= 32) {
    work_path_start_index = wp_l;

#if 0    
    // include parent directory
    for (int i=wp_l-1; i >= 0; --i) {
      if (wp_s[i] == '/') {
        work_path_start_index = i;
        break;
      }
    }
#endif
    
  }
  
  vfd_update();
}

bool Interface::on_tick(int secs)
{
  return tick % (secs*update_hz*2) < secs*update_hz;
}

void Interface::vfd_update()
{
  string text;

  if (playing) {

    // play mode
    
    if (mpd_status.song_pos < 0 ||
        (int)songs.size() <= mpd_status.song_pos)
      return;

    // line 1: animated file name
    string line1 = songs[mpd_status.song_pos];
    boost::replace_all(line1, ".mp3", "");
    line1 += "  ***  ";

    if (freeze_ticks-- > 0) {
      
#if 0
      int found = 0;
      for (int i=line1.size()-1; i >= 0; --i) {
        if (line1[i] == '/') {
          ++found;
          if (found >= 2) {
            delta_tick1 = tick - i;
            break;
          }
        }
      }
#endif
      
      delta_tick2 = tick;
    }

    string disp1;
    for (int i=0; i < 16; i++) {
      disp1 += line1[(tick-delta_tick1+i) % line1.size()];
    }

    string line2;
    
    if (pausing) {

      line2 = "Paused          ";

    } else {
      
      // line 2: 1/153 xx:xx xx:xx 320
      line2 = (boost::format("%d/%d %d:%02d %d:%02d %d ")
               % (mpd_status.song_pos+1)
               % mpd_status.queue_len
               % (mpd_status.elapsed_time / 60)
               % (mpd_status.elapsed_time % 60)
               % (mpd_status.total_time / 60)
               % (mpd_status.total_time % 60)
               % mpd_status.bit_rate).str();
    }

    string disp2;
    for (int i=0; i < 16; i++) {
      disp2 += line2[((tick-delta_tick2)/2+i) % line2.size()];
    }
        
    // swap lines
    if (on_tick(SWAP_SECS)) swap(disp1, disp2);
    text = disp1 + disp2;

  } else {

    // selection mode

    if (tick - last_action_tick >= SAVER_SECS * update_hz) {

      // screen saver
      text = "                                ";
      text[(tick - last_action_tick) % 32] = '*';

    } else if (mode2) {

      // select play mode

      text = "Play Mode:      " + play_mode.to_string();

    } else {

      // navigate through tree
      
      int ci = current_item + 1;
      if (!work_item_nodes.size())
        ci = 0;
      
      // track/total
      text = (boost::format("%d/%d ")
              % ci
              % work_item_nodes.size()
              ).str();
      
      // work path without home_path
      string wp;
      
      if (work_item_nodes.size())
        wp = work_item_strs[current_item];
      else
        wp = work_node->path;

      string title = wp;
      
      if (work_path_start_index > 0) {
        // shorten path if too long
        title = "..." + wp.substr(work_path_start_index);
      }

      int ts = title.size();

      if (ts + text.size() < 32) {

        text += title;
        selection_tick = tick;

      } else {

        for (int i=0; text.size() < 32; ++i) {
          int r = (i + tick - selection_tick + ts) % ts;
          text += title[r];
        }
      }
    }
  }

  vfd(text);
}
