#include "PlayLists.H"

#include <algorithm>
#include <boost/format.hpp>
#include <boost/thread.hpp>
#include <boost/filesystem.hpp>
#include <fstream>

#include "MP3Info.H"
#include "PlayNode.H"
#include "PlayMode.H"

using namespace std;
using namespace boost::filesystem;

typedef vector<PlayNode*> NodeVec;
typedef vector<NodeVec> SeqVec;

// stores all nodes in sequence vector in a single vector
static void flatten(const SeqVec &sv, NodeVec &v)
{
  v.clear();
  
  FORALL (sv, i) {
    FORALL (*i, j) {
      v.push_back(*j);
    }
  }
}

// collect marked files, shuffle or keep order
// adds new node sequences to sv
static void tree2list(PlayNode *node, RNG &rng, bool shuffle, SeqVec &sv)
{
  if (node->directory) {

    // directory

    // collect successors
    
    SeqVec sv2;

    FORALL (node->succ, i) {
      if (!node->marked) continue;
      tree2list(*i, rng, shuffle, sv2);
    }
    
    if (shuffle && node->random) {

      // randomize and append
      std::shuffle(sv2.begin(), sv2.end(), rng);

      FORALL (sv2, j) {
        sv.push_back(*j);
      }
      
    } else {

      // flatten and append

      NodeVec v;
      flatten(sv2, v);

      // add fixed sequence
      sv.push_back(v);
    }

  } else {

    // file

    // add singleton    
    NodeVec v;
    v.push_back(node);
    sv.push_back(v);
  }
}



// read all lists from directory
int PlayLists::read(const string &dir)
{
  cout << "read lists from " << dir << endl;
    
  path p(dir);

  lists.clear();
    
  if (!exists(p) || !is_directory(p))
    ERR("can't read playlist directory: " << dir);

  typedef vector<path> vec;             // store paths,
  vec v;                                // so we can sort them later
    
  copy(directory_iterator(p), directory_iterator(), back_inserter(v));
    
  sort(v.begin(), v.end());             // sort, since directory iteration

  for (auto it = v.begin(), end = v.end(); it != end; ++it) {

    if (is_regular_file(*it)) {

      string full_path = it->string();

      if (full_path.substr(full_path.size()-4) != ".spl")
        continue;

      string display_name = it->filename().string();
      display_name = display_name.substr(0, display_name.size()-4);
        
      cout << "read list: " << display_name << " ..." << flush;
      int res = add_list(display_name, full_path);

      if (res <  0) ERR("error reading list " << full_path);
      if (res == 0) ERR("empty reading list " << full_path);
      cout << " done" << endl;
    }
  }
  return lists.size();
}

void PlayLists::create_song_list(int index, const PlayNode *tree, RNG &rng, bool shuffle,
                                 const PlayMode &play_mode,
                                 string &display_name, vector<string> &songs)
{
  if (index < 0 || index >= (int)lists.size()) ERR("index out of range " << index);

  PlayList &play_list = lists[index];
  
  display_name = play_list.display_name;

  songs.clear();

  SeqVec all;
  
  FORALL (play_list.list, it) {

    PlayNode *root = tree->find(*it);
    if (!root) {
      cerr << "can't find list item " << *it << endl;
      continue;
    }
    
    SeqVec sv;
    tree2list(root, rng, shuffle, sv);

    all.insert(all.end(), sv.begin(), sv.end());
  }

  if (shuffle)
    std::shuffle(all.begin(), all.end(), rng);

  NodeVec nv;
  flatten(all, nv);
    
  FORALL (nv, it) {
    songs.push_back((*it)->path);
  }
}

// add a list of songs with name
void PlayLists::add_list(const std::string &display_name, const std::vector<string> &songs)
{
  PlayList playlist;
  playlist.display_name = display_name;
  playlist.path = "";

  FORALL (songs, it) {
    playlist.list.push_back(*it);
  }

  lists.push_back(playlist);
}


// >=0: ok (new list size), <0: error
int PlayLists::add_list(const string &display_name, const string &full_path)
{
  ifstream is(full_path);
  if (!is) { is.close(); return -1; }

  PlayList playlist;
  playlist.display_name = display_name;
  playlist.path = full_path;

  const int MAX_LEN = 1000;
  char s[MAX_LEN+1];
    
  for (;;) {
    is.getline(s, MAX_LEN);
    if (is.eof()) break;
    if (!is) ERR("read error " << full_path);
    playlist.list.push_back(string(s));
  }
  is.close();
  lists.push_back(playlist);
  return playlist.list.size();
}
