#include <fstream>
#include <iostream>
#include <stdio.h>   /* Standard input/output definitions */
#include <string.h>  /* String function definitions */
#include <unistd.h>  /* UNIX standard function definitions */
#include <fcntl.h>   /* File control definitions */
#include <errno.h>   /* Error number definitions */
#include <termios.h> /* POSIX terminal control definitions */
#include <stdlib.h>
#include <vector>
#include "CPP.H"
#include "Serial.H"

using namespace std;

static const char* dev_monitor  = "/dev/serial/by-id/usb-FTDI_USB_Serial_Converter_FTG6EIH4-if00-port0";
static const char* dev_receiver = "/dev/serial/by-id/usb-FTDI_USB_Serial_Converter_FTG8TWAI-if00-port0";

struct Command {
  string cmd;
  string serial_cmd;
};


static vector<Command> mon_commands = {
  { "on",        "PON"     },
  { "off",       "POF"     },
  { "input-1",   "IMS:SL1" },    
  { "input-2",   "IMS:SL2" },
  { "input-3",   "IMS:SL3" },
  { "input-pc",  "IMS:PC1" }
};

static vector<Command> rec_commands = {
  { "on",             "PWR01"    },
  { "off",            "PWR00"    },
  { "mute",           "AMT01"    },
  { "unmute",         "AMT00"    },
  { "mute-toggle",    "AMTTG"    },
  { "volume+",        "MVLUP1"   },
  { "volume-",        "MVLDOWN1" },
  { "input-bluray",   "SLI10"    },    
  { "input-pc-audio", "SLI00"    },
  { "input-pop",      "SLI01"    },
  { "input-game",     "SLI02"    },
  { "input-pc-video", "SLI05"    },
  { "input-aux",      "SLI03"    },
  { "input-classic",  "SLI23"    }
};


void Serial::write_mon_commands(ostream &os)
{
  FORALL (mon_commands, it) {
    os << it->cmd << " ";
  }
}

void Serial::write_rec_commands(ostream &os)
{
  FORALL (rec_commands, it) {
    os << it->cmd << " ";
  }
}


void Serial::write_macro_commands(std::ostream &os)
{
  os << "bluray video audio game classic pop off ";
}


static int open_port(const char *device)
{
  int fd = open(device, O_RDWR | O_NOCTTY | O_NDELAY);
  if (fd < 0) return fd;
  
  fcntl(fd, F_SETFL, 0); // blocking ?

  struct termios options;

  // Get the current options for the port...
  tcgetattr(fd, &options);

  // Set the baud rates to 9600
  cfsetispeed(&options, B9600);
  cfsetospeed(&options, B9600);

  // Enable the receiver and set local mode...
  options.c_cflag |= (CLOCAL | CREAD);

  // No parity (8N1)
  options.c_cflag &= ~PARENB;
  options.c_cflag &= ~CSTOPB;
  options.c_cflag &= ~CSIZE;
  options.c_cflag |= CS8;

  // Set the new options for the port...
  tcsetattr(fd, TCSANOW, &options);

  return fd;
}


// monitor

// != 0 <=> error
int Serial::mon_cmd(const string &cmd)
{  
  const char *device = dev_monitor;
  const char c_start = 2;
  const char c_end = 3;

  /* plasma:

     c_start msg c_end + wait for reply
     => c_start response c_end
     
     PON      turn on      => PON
     POF      turn off     => POF
     IMS:SL1  input slot 1 => IMS
     IMS:SL2  input slot 2 => IMS
     IMS:SL3  input slot 3 => IMS
     IMS:PC1  input pc     => IMS
     
   */
  
  string serial_cmd;
  
  FORALL (mon_commands, it) {
    if (it->cmd == cmd) {
      serial_cmd = it->serial_cmd;
      break;
    }
  }

  if (serial_cmd.empty()) {
    cerr << "unknown command " << cmd << endl;
    return 1;
  }

  cerr << "execute monitor command " << cmd << endl;
  
  int fd = open_port(device);
  if (fd < 0) {
    cerr << "open_port: unable to open " << device << endl;
    return 1;
  }

  const int BUF_LEN = 1024;
  char buffer[BUF_LEN];
  snprintf(buffer, BUF_LEN-1, "%c%s%c", c_start, serial_cmd.c_str(), c_end);
  
  if (write(fd, buffer, strlen(buffer)) != (int)strlen(buffer)) {
    cerr << "write error" << endl;
    close(fd);
    return 1;
  }

  return 0;
}
 

// receiver

// != 0 <=> error
int Serial::rec_cmd(const string &cmd)
{
  const char *device = dev_receiver;
  const char *s_start = "!1";
  const char c_end   = '\n';

  /* codes

     c_start msg c_end + wait for reply
     => c_start response c_end

     cmd      function           response (ignored)

     PWR01    turn on            PWR01
     PWR00    turn off           PWR00
     PWRQSTN  power level        PWR0x

     AMT01    mute on            AMT01
     AMT00    mute off           AMT00
     AMTTG    mute toggle        AMT0x
     AMTQSTN  mute level         AMT0x

     MVLUP    volume up 0.5      MVLxx  (integer volume level xx)
     MVLDOWN  volume down 0.5    -"-
     MVLUP1   volume up 1        -"-
     MVLDOWN1 volume down 1      -"-
     MVLQSTN  volume state       -"-

     SLI00    VCR/DVR (PC Music)  LMDyy SLIxx SLIxx or only one SLIxx if there is no change
     SLI01    CBL/SAT
     SLI02    GAME
     SLI03    AUX
     SLI05    PC (Video)
     SLI10    BluRay
     SLI22    Phono
     SLI23    TV/CD
     SLI24    Tuner FM
     SLI25    Tuner AM
     SLI26    Tuner
     SLI27    Net
     SLI28    My Favorites
     SLI29    USB(F)
     SLI2A    USB(R)
     SLI2B    Net Internet Radio
     SLIQSTN
  */

  string serial_cmd;
  
  FORALL (rec_commands, it) {
    if (it->cmd == cmd) {
      serial_cmd = it->serial_cmd;
      break;
    }
  }

  if (serial_cmd.empty()) {
    cerr << "unknown command " << cmd << endl;
    return 1;
  }

  cerr << "execute receiver command " << cmd << endl;

  int fd = open_port(device);
  if (fd < 0) {
    cerr << "open_port: unable to open " << device << endl;
    return 1;
  }

  const int BUF_LEN = 1024;
  char buffer[BUF_LEN];
  snprintf(buffer, BUF_LEN-1, "%s%s%c", s_start, serial_cmd.c_str(), c_end);
  
  if (write(fd, buffer, strlen(buffer)) != (int)strlen(buffer)) {
    cerr << "write error" << endl;
    close(fd);
    return 1;
  }

#if 0

  // wait for response
  // hung once in tests, not sure why
  // there also may be interence from remote control
  // as its command responses are also sent to the tty
  
  int match = 0, other = 0;
  int i;

  for (;;) {

    // read one message

    i = 0;

    for (;;) {
      if (i >= BUF_LEN-1) {
        cerr << "message too long" << endl; exit(0);
      }
      if (read(fd, &buffer[i], 1) != 1) {
        cerr << "read error" << endl; exit(0);
      }
      cout << (int)buffer[i] << " " << buffer[i] << endl;
      if (buffer[i] == 26) break;
      i++;
    } 
    if (buffer[0] != '!') {
      cerr << "message corrupt 1 " << endl; exit(0);
    }

    // !1<CMD> ?
    if (serial_cmd[0] == buffer[2] &&
        serial_cmd[1] == buffer[3] &&
        serial_cmd[2] == buffer[4])
      ++match;
    else
      ++other;

    if (match == 1 && other == 0) break;
    if (match == 2 && other > 0)  break;
  }

  if (i < 2) {
    cerr << "index corrupt" << endl; exit(0);
  }

  int res = 0;
  buffer[i] = 0;
  if (sscanf(&buffer[i-2], "%x", &res) != 1) {
    cerr << "result not a hex number" << endl;
    exit(0);
  }
  cout << res << endl;
#endif

  close(fd);
  return 0;
}


// != 0: error
int Serial::macro(const std::string &cmd)
{
  cerr << "execute macro " << cmd << endl;
  
  if (cmd == "bluray") {   // turn on bluray player
    mon_cmd("on");
    rec_cmd("on");
    sleep(1);
    mon_cmd("input-1");
    rec_cmd("input-bluray");
    
  } else if (cmd == "video") { // watch video on pc
    mon_cmd("on");
    rec_cmd("on");
    sleep(1);
    mon_cmd("input-pc");      
    rec_cmd("input-pc-video");
      
  } else if (cmd == "audio") { // listen to pc audio 
    mon_cmd("off");
    rec_cmd("on");
    sleep(1);
    rec_cmd("input-pc-audio");

  } else if (cmd == "game") { // wii
    mon_cmd("on");
    rec_cmd("on");
    sleep(1);
    mon_cmd("input-2");
    rec_cmd("input-game");

  } else if (cmd == "classic") { // classic music cd player
    mon_cmd("off");
    rec_cmd("on");
    sleep(1);
    rec_cmd("input-classic");

  } else if (cmd == "pop") { // pop cd player
    mon_cmd("off");
    rec_cmd("on");
    sleep(1);
    rec_cmd("input-pop");

  } else if (cmd == "off") { // all off
    mon_cmd("off");
    rec_cmd("off");

  } else {
    cerr << "unknown macro " << cmd << endl;
    return 1;
  }
  
  return 0;
}

