// based on:

/*
 * libwebsockets-test-server - libwebsockets test implementation
 *
 * Copyright (C) 2010-2011 Andy Green <andy@warmcat.com>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation:
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *  MA  02110-1301  USA
 */

#include "WebServer.H"
#include "libwebsockets/libwebsockets.h"

#include <cstdio>
#include <cstdlib>
#include <unistd.h>
#include <getopt.h>
#include <cstring>
#include <sys/time.h>

using namespace std;

static bool debug = true;

// give callback functions access to server
std::map<libwebsocket_context*, WebServer*> context2server;

// get server from context
WebServer& get_server(libwebsocket_context *context)
{
  FIND(context2server, it, context);
  if (it == context2server.end()) ERR("server not found");
  return *it->second;
}

WebServer::WebServer(int port)
  : port(port)
{
}

WebServer::~WebServer()
{
}

#if 0
struct WriteThread
{
  WriteThread(ClientData& cd) : cd(cd) { }
  
  void operator()()
  {
    for (;;) {
      if (done) return;
      std::unique_lock<std::mutex> lock(cd.mutex);
      cd.cond_var.wait(lock);

      while (!queue.empty()) {

        msg = cd.queue.front();
        cd.queue.pop();
        
        const int N = 1024;
        char buf[LWS_SEND_BUFFER_PRE_PADDING + N +
                 LWS_SEND_BUFFER_POST_PADDING];
        
        size_t len = msg.size();
        if (len >= N-1) len = N-1;
        
        strncpy(&buf[LWS_SEND_BUFFER_PRE_PADDING], msg.c_str(), N-1);
        buf[LWS_SEND_BUFFER_PRE_PADDING + N - 1] = 0;

        libwebsocket_write(wsi, (unsigned char*)&buf[LWS_SEND_BUFFER_PRE_PADDING], len+1,
                           &protocols[PROTOCOL_JOX]);
      }
    }
  }

  ClientData &cd;
  bool done = false;
}

#endif
  

void WebServer::add_client(libwebsocket_context* context, libwebsocket *wsi)
{
  std::lock_guard<std::mutex> lock(clients_mutex);
  ClientData& cd = clients[wsi] = ClientData();
  cd.context = context;
  cd.wsi = wsi;
}

void WebServer::remove_client(libwebsocket *wsi)
{
  std::lock_guard<std::mutex> lock(clients_mutex);
  clients.erase(wsi);
}

bool WebServer::get_msg(libwebsocket *wsi, string &msg)
{
  std::lock_guard<std::mutex> lock(clients_mutex);
  ClientData& cd = clients[wsi];
  if (!cd.queue.empty()) {
    msg = cd.queue.front();
    cd.queue.pop();
    return true;
  }
  return false;
}

static int callback_http(struct libwebsocket_context *context,
                         struct libwebsocket *wsi,
                         enum libwebsocket_callback_reasons reason, void *user,
                         void *in, size_t len);

static int callback_jox(struct libwebsocket_context *context,
                        struct libwebsocket *wsi,
                        enum libwebsocket_callback_reasons reason, void *user,
                        void *in, size_t len);

/* jox protocol */

/*
 * one of these is auto-created for each connection and a pointer to the
 * appropriate instance is passed to the callback in the user parameter
 *
 * for this example protocol we use it to individualize the count for each
 * connection.
 */

struct per_session_data_jox
{
  int dummy;
};

/*
 * This demo server shows how to use libwebsockets for one or more
 * websocket protocols in the same server
 *
 * It defines the following websocket protocols:
 *
 *  jox-protocol: broadcast jox state to clients,
 *                receive jox commands
 */

enum demp_protocols {
/* always first */
  PROTOCOL_HTTP = 0,
  PROTOCOL_JOX
};

static struct libwebsocket_protocols protocols[] = {
  // first protocol must always be HTTP handler

  {
    "http-only",	// name
    callback_http,	// callback
    0			// per_session_data_size
  },

  {
    "jox-protocol",
    callback_jox,
    sizeof(struct per_session_data_jox),
  },

  {
    nullptr, nullptr, 0	// End of list
  }
};


static int callback_http(struct libwebsocket_context *context,
                         struct libwebsocket *wsi,
                         enum libwebsocket_callback_reasons reason, void *user,
                         void *in, size_t len);

// list of supported protocols and callbacks

int WebServer::run(Handler *new_handler)
{
  handler = new_handler;
  
  const char *cert_path = "./libwebsockets-test-server.pem";
  const char *key_path =  "./libwebsockets-test-server.key.pem";
  int use_ssl = 0;
  struct libwebsocket_context *context;
  int opts = 0;

  // use_ssl = 1;
  // opts = LWS_SERVER_OPTION_DEFEAT_CLIENT_MASK;

  if (!use_ssl)
    cert_path = key_path = nullptr;

  context = libwebsocket_create_context(port, "interface", protocols,
                                        libwebsocket_internal_extensions,
                                        cert_path, key_path, -1, -1, opts);
  if (context == nullptr) {
    cerr << "libwebsocket init failed" << endl;
    return -1;
  }

  context2server[context] = this;
  
  int n = 0;
  while (!done && n >= 0) {

    /*
     * This example server does not fork or create a thread for
     * websocket service, it all runs in this single loop.  So,
     * we have to give the websockets an opportunity to service
     * "manually".
     *
     * If no socket is needing service, the call below returns
     * immediately and quickly.  Negative return means we are
     * in process of closing
     */

    n = libwebsocket_service(context, 50);

#if 0    
    for (;;) {
      string msg;
      bool work = false;
      {
        std::lock_guard<std::mutex> lock(queue_mutex);
        if (!queue.empty()) {
          msg = queue.front();
          queue.pop();
          work = true;
        }
      }
      if (!work) break;
      
      const int N = 1024;
      char buf[LWS_SEND_BUFFER_PRE_PADDING + N +
               LWS_SEND_BUFFER_POST_PADDING];
      
      size_t len = msg.size();
      if (len >= N-1) len = N-1;
      
      strncpy(&buf[LWS_SEND_BUFFER_PRE_PADDING], msg.c_str(), N-1);
      buf[LWS_SEND_BUFFER_PRE_PADDING + N - 1] = 0;
      libwebsockets_broadcast(&protocols[PROTOCOL_JOX],
                              (unsigned char*)&buf[LWS_SEND_BUFFER_PRE_PADDING], len+1); // text + 0
    }
#endif    
  }

  libwebsocket_context_destroy(context);
  context2server.erase(context);
  return 0;
}

// add message to each client's queue
void WebServer::broadcast(const std::string &msg)
{
  std::lock_guard<std::mutex> lock(clients_mutex);

  FORALL (clients, it) {
    ClientData &cd = it->second;
    if (cd.queue.size() >= MAX_QUEUE_SIZE) {
      // reset queue if too long
      while (!cd.queue.empty())
        cd.queue.pop();
      cerr << "RESET QUEUE" << endl;
    }
    cd.queue.push(msg);
    cerr << "on_writable_1 " << it->second.wsi << endl;
    libwebsocket_callback_on_writable(it->second.context, it->second.wsi);
  }
}

// was: #define LOCAL_RESOURCE_PATH INSTALL_DATADIR"/libwebsockets-test-server"

#define LOCAL_RESOURCE_PATH "."

/* this protocol server (always the first one) just knows how to do HTTP */

static int callback_http(struct libwebsocket_context *context,
                         struct libwebsocket *wsi,
                         enum libwebsocket_callback_reasons reason, void *user,
                         void *in, size_t len)
{
  // char client_name[128];
  // char client_ip[128];

  switch (reason) {
  case LWS_CALLBACK_HTTP:

    fprintf(stderr, "serving HTTP URI %s\n", (char *)in);

    if (in && strcmp((char*)in, "/favicon.ico") == 0) {
      if (libwebsockets_serve_http_file(wsi,
                                        LOCAL_RESOURCE_PATH"/favicon.ico", "image/x-icon"))
        cerr << "Failed to send favicon" << endl;
      break;
    }

    // send the script... when it runs it'll start websockets

    if (libwebsockets_serve_http_file(wsi,
                                      LOCAL_RESOURCE_PATH"/jox.html", "text/html"))
      cerr << "Failed to send HTTP file" << endl;
    break;

  default:
    break;
  }

  return 0;
}



static int callback_jox(struct libwebsocket_context *context,
                        struct libwebsocket *wsi,
                        enum libwebsocket_callback_reasons reason,
                        void *user, void *in, size_t len)
{
  //int n;
  // struct per_session_data_jox *psd = (per_session_data_jox*) user;

  if (debug)
    cerr << "CALLBACK " << reason << " " << wsi << " " << user << endl;

  switch (reason) {

  case LWS_CALLBACK_ESTABLISHED:
    {
      if (debug) cerr << "callback_jox: LWS_CALLBACK_ESTABLISHED " << wsi << endl;
      // add client
      get_server(context).add_client(context, wsi);
    }
    break;

  case LWS_CALLBACK_CLOSED:
    {
      if (debug) cerr << "callback_jox: LWS_CALLBACK_CLOSED " << wsi << endl;
      // remove client
      get_server(context).remove_client(wsi);
    }
    break;

  case LWS_CALLBACK_CLIENT_CONNECTION_ERROR:
    if (debug) cerr << "callback_jox: LWS_CALLBACK_CLIENT_CONNECTION_ERROR " << wsi << endl;
    break;

  case LWS_CALLBACK_SERVER_WRITEABLE:
    {
      // non-blocking write possible => send message

      string msg;
      if (get_server(context).get_msg(wsi, msg)) {
        
        if (debug)
          cerr << "callback_jox: WRITABLE" << endl;

        const int N = 1024;
        char buf[LWS_SEND_BUFFER_PRE_PADDING + N +
                 LWS_SEND_BUFFER_POST_PADDING];
      
        size_t len = msg.size();
        if (len >= N-1) len = N-1; // truncate if necessary
      
        strncpy(&buf[LWS_SEND_BUFFER_PRE_PADDING], msg.c_str(), N-1);
        buf[LWS_SEND_BUFFER_PRE_PADDING + N - 1] = 0;
      
        libwebsocket_write(wsi, (unsigned char*)&buf[LWS_SEND_BUFFER_PRE_PADDING], len+1,  LWS_WRITE_TEXT);
        // request writeable callback for possible follow-up messages
        int n = libwebsocket_callback_on_writable(context, wsi);
        if (n < 0) {
          cerr << "ERROR writing to socket" << endl;
          return 1;
        }
      }
    }
    break;
    
  case LWS_CALLBACK_BROADCAST:
    ERR("broadcast may hang");
    break;

  case LWS_CALLBACK_RECEIVE:
    cerr << "received " << (char*)in << endl;
    get_server(context).get_handler()->handle((char*)in);
    // code from test_server.c - not sure what this does
    cerr << "on_writable_2 " << wsi << " " << user << endl;
    libwebsocket_callback_on_writable_all_protocol(libwebsockets_get_protocol(wsi));
    break;

  default:
    break;
  }

  return 0;
}
