/*
 *
 * jox - "jukebox" - mp3 player connected to mpd, lirc, vfd
 * recent additions:
 *  - web interface
 *  - managing receiver and monitor via RS232
 *
 * (c) 2012 Michael Buro
 *
 */

#include <cerrno>
#include <unistd.h>
#include <cstdarg>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <getopt.h>
#include <signal.h>
#include <lirc/lirc_client.h>
#include <mpd/client.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <iostream>
#include <fstream>
#include <random>
#include <vector>
#include <algorithm>
#include <boost/format.hpp>
#include <boost/thread.hpp>
#include <boost/filesystem.hpp>
#include <curses.h>
#include <mutex>
#include <thread>

#undef ERR

#include "PlayLists.H"
#include "Interface.H"
#include "Term.H"
#include "MPD.H"
#include "WebServer.H"
#include "Serial.H"

using namespace std;
using namespace boost::filesystem;

const string VERSION = "1.0";
const int UPDATE_HZ = 5;       // for VFD and web updates
const int INITIAL_PLAY_MODE = PlayMode::MODE_GOOD; // audio

lirc_config *lirc_conf = 0;
mpd_connection *mpd_conn = 0;
const char *mpd_host = "localhost";
int mpd_port = 6600;
int web_port = 9000;
ostream *vfd_os = 0;
WebServer *web_server = 0;
const string home_dir = "music"; // link to MPD music folder
bool term_set = false; // true => reset terminal
bool lirc_inited = false;
std::thread *timer_thread = 0;
std::thread *web_thread = 0;

bool done = false; // true => finish, program about to exit

std::random_device rd;
RNG rng(rd());

std::mutex interface_mutex; // only one thread in interface functions

// free resources, clear display, reset terminal, and exit

void cleanup_and_exit(int ret)
{
  // wait to become sole interface user
  std::lock_guard<std::mutex> lock(interface_mutex);
      
  cout << "cleanup" << endl;

  // reset terminal
  if (term_set) {
    reset_term();
  }

  // free lirc
  if (lirc_conf)
    lirc_freeconfig(lirc_conf);

  if (lirc_inited)
    lirc_deinit();

  // clear VFD
  if (vfd_os) {
    (*vfd_os) << " ";
    (*vfd_os).flush();
    // (*vfd_os).close(); done automatically
  }
  
  // disconnect from MPD
  if (mpd_conn) {
    mpd_stop(mpd_conn);
    mpd_clear(mpd_conn);
    mpd_connection_free(mpd_conn);
  }

  // free web server
  delete web_server;
  
  exit(ret);
}


void sighandler(int sig)
{
  done = true;
}

struct TimerThread
{
  Interface &interface;
  
  TimerThread(Interface &interface) : interface(interface) { }

  void operator()()
  {
    int tick = 0;
    while (!done) {
      {
        std::lock_guard<std::mutex> lock(interface_mutex);
        interface.update(tick++);
      }
      usleep(1000000/UPDATE_HZ);

      if (web_server) {
        string status;
        {
          std::lock_guard<std::mutex> lock(interface_mutex);
          status = interface.status();
        }
        web_server->broadcast(status);
      }
    }

    cleanup_and_exit(0);
  }
};


// audio commands
static void execute_cmd(Interface &interface, const string &cmd)
{
  std::lock_guard<std::mutex> lock(interface_mutex);
  
  cout << "execute " << cmd << endl;

  bool executed = true;
  
  if (cmd == "playpause") {
    
    interface.action_playpause();
    
  } else if (cmd == "stop") {
    
    interface.action_stop();          
    
  } else if (cmd == "previous") {
    
    interface.action_previous();
    
  } else if (cmd == "next") {
    
    interface.action_next();          
    
  } else if (cmd == "cw") {
    
    interface.action_cw();
    
  } else if (cmd == "ccw") {
    
    interface.action_ccw();
    
  } else if (cmd == "select") {
    
    interface.action_select();
    
  } else if (cmd == "seek+") {

    interface.action_seek_forward();
              
  } else if (cmd == "seek-") {

    interface.action_seek_backward();
              
  } else if (cmd == "mode") {

    interface.action_mode();
              
  } else if (cmd == "exit") {

    done = true; // cleanup and exit
              
  } else {
    
    cerr  << "command " << cmd << " not recognized" << endl;
    executed = false;
  }

  if (executed) {
    interface.set_mode(Interface::Mode::AUDIO);
  }
}

// kill all xine instances
static void kill_video()
{
  system("killall -9 xine");
}

// stop audio playback and set new mode
static void synchronized_stop(Interface &interface, Interface::Mode mode)
{
  std::lock_guard<std::mutex> lock(interface_mutex);
  interface.set_mode(mode);
  interface.action_stop();
}


struct WebThread : public WebServer::Handler
{
  WebServer &web_server;
  Interface &interface;
  
  WebThread(WebServer &web_server, Interface &interface)
    : web_server(web_server), interface(interface) { }

  void operator()()
  {
    web_server.run(this);
  }

  void handle(const std::string &msg) {
    cout << endl << "handle: " << msg << endl;

    // volume
    if      (msg == "volume+")     Serial::rec_cmd("volume+");
    else if (msg == "volume-")     Serial::rec_cmd("volume-");
    else if (msg == "mute-toggle") Serial::rec_cmd("mute-toggle");

    // macros
    else if (msg == "off") {
      Serial::macro("off");
      synchronized_stop(interface, Interface::Mode::OFF);
      kill_video();

    } else if (msg == "bluray") {
      synchronized_stop(interface, Interface::Mode::BLU);
      kill_video();
      Serial::macro("bluray");    
      
    } else if (msg == "video") {
      synchronized_stop(interface, Interface::Mode::VIDEO);
      Serial::macro("video");

    } else if (msg == "audio") {
      synchronized_stop(interface, Interface::Mode::AUDIO);
      kill_video();
      Serial::macro("audio");    

    } else if (msg == "game") {
      synchronized_stop(interface, Interface::Mode::WII);
      kill_video();
      Serial::macro("game");    

    } else if (msg == "classic") {
      synchronized_stop(interface, Interface::Mode::CD_CLASSIC);
      kill_video();
      Serial::macro("classic");

    } else if (msg == "pop") {
      synchronized_stop(interface, Interface::Mode::CD_POP);      
      kill_video();
      Serial::macro("pop");

    } else if (msg == "monitor-on") {
      Serial::mon_cmd("on");
      sleep(1);
      Serial::mon_cmd("input-pc");      

    } else if (msg == "monitor-off") {
      Serial::mon_cmd("off");

    } else if (msg == "tagesschau") {
      kill_video();
      system("irsend SIMULATE \"000000008001000e 00 KEY_BACKSPACE remote\"");
      
    } else if (msg == "bbc") {
      kill_video();
      system("irsend SIMULATE \"0000000080010189 00 KEY_VIDEO remote\"");

    } else if (msg == "daily-show") {
      kill_video();
      system("irsend SIMULATE \"0000000080010188 00 KEY_AUDIO remote\"");
      
    } else if (msg == "dvd") {
      kill_video();
      system("irsend SIMULATE \"0000000080010185 00 KEY_DVD remote\"");

    } else if (msg == "tv") {
      kill_video();
      system("irsend SIMULATE \"0000000080010179 00 KEY_TV remote\"");

    } else if (msg == "dvd-menu") {
      system("irsend SIMULATE \"000000008001008b 00 KEY_MENU remote\"");

    } else if (msg == "dvd-left") {
      system("irsend SIMULATE \"0000000080010069 00 KEY_LEFT remote\"");

    } else if (msg == "dvd-up") {
      system("irsend SIMULATE \"0000000080010067 00 KEY_UP remote\"");

    } else if (msg == "dvd-select") {
      system("irsend SIMULATE \"000000008001001c 00 KEY_ENTER remote\"");

    } else if (msg == "dvd-down") {
      system("irsend SIMULATE \"000000008001006c 00 KEY_DOWN remote\"");

    } else if (msg == "dvd-right") {
      system("irsend SIMULATE \"000000008001006a 00 KEY_RIGHT remote\"");

    } else if (msg == "video-zoom-in") {
      system("irsend SIMULATE \"0000000080010073 00 KEY_VOLUMEUP remote\"");

    } else if (msg == "video-zoom-out") {
      system("irsend SIMULATE \"0000000080010072 00 KEY_VOLUMEDOWN remote\"");

    } else if (msg == "video-exit") {
      kill_video();
      
    } else if (msg == "video-playpause") {
      system("irsend SIMULATE \"0000000080010077 00 KEY_PAUSE remote\"");

    } else if (msg == "video-seek-") {
      system("irsend SIMULATE \"000000008001019c 00 KEY_PREVIOUS remote\"");

    } else if (msg == "video-seek+") {
      system("irsend SIMULATE \"0000000080010197 00 KEY_NEXT remote\"");

    } else if (msg == "video-chapter-") {
      system("irsend SIMULATE \"0000000080010193 00 KEY_CHANNELDOWN remote\"");

    } else if (msg == "video-chapter+") {
      system("irsend SIMULATE \"0000000080010192 00 KEY_CHANNELUP remote\"");

    } else {
      execute_cmd(interface, msg);
    }
  }
};



int main(int argc, char *argv[])
{
  int daemonize = 0;  // not suppported used
  const char *program = "jox";
  bool vfd_available = true;
  bool lirc_available = true;
  bool mpd_available = true;  
  bool web_server_available = true;
  
  struct option long_options[] = {
    {"help",     no_argument,       0, 'h'},
    {"version",  no_argument,       0, 'v'},
    {"daemon",   no_argument,       0, 'd'},
    {"name",     required_argument, 0, 'n'},
    {"novfd",    no_argument,       0, 'V'},
    {"nolirc",   no_argument,       0, 'L'},
    {"nompd",    no_argument,       0, 'M'},      
    {"noweb",    no_argument,       0, 'W'},    
    {0, 0, 0, 0}
  };

  for (;;) {

    int c;
    c = getopt_long(argc, argv, "hvdn:VLMW", long_options, NULL);
    if (c == -1) break;

    switch (c) {

    case 'h':
    help_msg:
      cout << "Usage: " << argv[0] <<  " [options]" << endl;
      cout << "-h --help    : display usage summary" << endl;
      cout << "-v --version : display version" << endl;
      cout << "-n --name    : use this program name" << endl;
      cout << "-V --novfd   : vfd unavailable - use stdout" << endl;
      cout << "-L --nolirc  : lirc unavailable - use stdin" << endl; 
      cout << "-M --nompd   : mpd unavailable - use stdout" << endl; 
      cout << "-W --noweb   : no webserver" << endl; 
      return EXIT_SUCCESS;
      
    case 'v':
      cout << program << " " << VERSION << endl;
      return EXIT_SUCCESS;

    case 'n':
      program = optarg;
      break;

    case 'V':
      vfd_available = false;
      break;

    case 'L':
      lirc_available = false;
      break;

    case 'M':
      mpd_available = false;
      break;

    case 'W':
      web_server_available = false;
      break;

    default:
      goto help_msg;
    }
  }

  if (optind < argc-1) {
    cerr << program << ": too many arguments" << endl;
    return EXIT_FAILURE;
  }

#if 0  
  if (optind >= argc) {
    cerr << "Expected argument after options" << endl;
    return EXIT_FAILURE;
  }
#endif

  //atexit(cleanup);

  // clean up when receiving signals
  signal(SIGABRT, &sighandler);
  signal(SIGTERM, &sighandler);
  signal(SIGINT,  &sighandler);

#if 0
  // test
  PlayLists play_lists(home_dir);

  int count = play_lists.read(home_dir + "/playlists");
  if (count <= 0) ERR("no playlist found");

  vector<string> songs;
  string display_name;
  play_lists.create_song_list(0, rng, true, display_name, songs);
  
  exit(0);
#endif
  
#if 0  

  // test
  // PlayNode *root = create_tree(path(string(getenv("HOME"))+"/music"));

  //PlayNode *root = create_tree(path("music/testA"));
  PlayNode *root = create_tree(path(home_dir + "/testD"));    

  if (!root) { cerr << "tree empty" << endl; exit(0); }
  
  SeqVec sv;
  tree2list(root, sv, true);
  NodeVec nv;
  flatten(sv, nv);

  FORALL (nv, i) {
    cout << (*i)->path << endl;
  }

  delete root;

  exit(0);
#endif
  
  // open VFD

  if (vfd_available) {
    vfd_os = new ofstream("/dev/lcd0");
    if (!(*vfd_os)) {
      cerr << "can't open VFD" << endl;
      cleanup_and_exit(EXIT_FAILURE);
    }

  } else 
    vfd_os = &cout;
  
  if (mpd_available) {
  
    // connect to mpd

    mpd_conn = mpd_connection_new(mpd_host, mpd_port, 0);
    
    if (mpd_conn == NULL) {
      cerr <<  "out of memory" << endl;
      cleanup_and_exit(EXIT_FAILURE);
    }
    
    if (mpd_connection_get_error(mpd_conn) != MPD_ERROR_SUCCESS) {
      cerr << "can't connect to mpd" << endl;
      cleanup_and_exit(EXIT_FAILURE);      
    }

    //if (options.password)
    // send_password(options.password, conn);

    cout << "connected to MPD. Updating ..." << flush;
    mpd_update(mpd_conn);
    cout << " done" << endl;
  }

  if (lirc_available) {
  
    // connect to lirc
  
    if (lirc_init((char*)program, daemonize ? 0 : 1) == -1) {
      cerr << "can't connect to LIRC" << endl;
      cleanup_and_exit(EXIT_FAILURE);
    }

    cout << "connected to LIRC" << endl;
    
    lirc_inited = true;
  }

  Serial::macro("audio"); // initialize receiver and monitor mode (has to match Interface::Mode:: below

  // music is a link to the MPD music folder
  Interface interface(home_dir, *vfd_os, mpd_conn, rng, UPDATE_HZ, Interface::Mode::AUDIO, INITIAL_PLAY_MODE);

  // start web server
  if (web_server_available) {
    web_server = new WebServer(web_port);
    web_thread = new std::thread(WebThread(*web_server, interface));
  }

  // start timer thread for updating interface regularly
  timer_thread = new std::thread(TimerThread(interface));

  if (lirc_inited) {
  
    cout << "LIRC INPUT" << endl;

    if (lirc_readconfig(0, &lirc_conf, 0) == 0) {

      char *c;
      int  ret;

      if (daemonize) {

        if (daemon(0, 0) == -1) {
          fprintf(stderr,"%s: can't daemonize\n", program);
          perror(program);
          exit(EXIT_FAILURE);
        }
      }
    
      char *code;

      while (lirc_nextcode(&code) == 0) {
      
        if (code == 0) continue;
        
        if (done) {
        free_and_break:;
          free(code);
          break;
        }
        
        while ((ret = lirc_code2char(lirc_conf, code, &c)) == 0 && c != 0) {

          if (done) goto free_and_break;
          
#ifndef NDEBUG
          if(!daemonize) {
            printf("Executing command \"%s\"\n", c);
          }
#endif
          execute_cmd(interface, c);

          free(code);
          if (ret == -1) break;
        }
      }
    }
    
  } else {

    // use keyboard for input

    cout << "KEYBOARD INPUT" << endl;

    set_term();
    term_set = true;
  
    FOREVER {
      int c = read_char();
      if (done)
        break;  // probably never get here - need to send signal to unblock?
      {
        std::lock_guard<std::mutex> lock(interface_mutex);
      
        if (c < 0) {
          switch (-c) {
          case KF_UP:
            interface.action_seek_forward();
            break;
          case KF_DOWN:
            interface.action_seek_backward();
            break;
          case KF_LEFT:
            interface.action_ccw();
            break;
          case KF_RIGHT:
            interface.action_cw();        
            break;
          }

        } else {

          switch (c) {
          case 'q':
            exit(0);
            break;
          case 'p':
            interface.action_playpause();
            break;
          case 's':
            interface.action_stop();
            break;
          case '\n':
            interface.action_select();
            break;
          case '[':
            interface.action_previous();
            break;
          case ']':
            interface.action_next();
            break;
          case '/':
            interface.action_mode();
            break;
          }
        }
      }
    }
  }

  if (timer_thread)
    timer_thread->join();

  if (web_thread)
    web_thread->join();

  return 0;
}
